from a6 import *
import pytest
import random


@pytest.fixture(scope="session", autouse=True)
def lexicon():
    with open('lexicon.txt', 'rt') as f:
        words = f.read().strip().split()
    return words, set(words)


def test_consistent_hash(lexicon):
    _, types = lexicon
    for w in types:
        assert cyclic_hash(w) == cyclic_hash(w)


def test_hash_size(lexicon):
    tokens, types = lexicon
    for w in types:
        assert cyclic_hash(w, size=8).bit_length() <= 8
    assert cyclic_hash(''.join(tokens), size=8).bit_length() <= 8


def test_hash_reverse(lexicon):
    tokens, types = lexicon
    collisions = 0
    for w in types:
        if cyclic_hash(w, size=32) == cyclic_hash(w[::-1], size=32):
            collisions += 1
    assert collisions / len(types) < 0.01


def test_hash_shuffled(lexicon):
    tokens, types = lexicon
    collisions = 0
    for w in types:
        wshuffle = w
        while w == wshuffle:
            wshuffle = list(w)
            random.shuffle(wshuffle)
            wshuffle = ''.join(wshuffle)
        if cyclic_hash(w, size=32) == cyclic_hash(wshuffle, size=32):
            collisions += 1
    assert collisions / len(types) < 0.01


def test_hash_substring(lexicon):
    tokens, types = lexicon
    collisions = 0
    for w in types:
        wlong = w + random.choice(w) * random.randint(1, 10)
        if cyclic_hash(w, size=32) == cyclic_hash(wlong, size=32):
            collisions += 1
    # print(collisions / len(types))
    assert collisions / len(types) < 0.01


def test_dedup_false_negatives(lexicon):
    tokens, types = lexicon
    nodups = []
    bf = BloomFilter()
    for w in sorted(tokens):
        if not bf.add(w):
            nodups.append(w)
    assert len(nodups) == len(set(nodups))  # no duplicates


def test_dedup_false_positives(lexicon):
    tokens, types = lexicon
    nodups = []
    bf = BloomFilter()
    for w in sorted(tokens):
        if not bf.add(w):
            nodups.append(w)
    nwords = len(types)

    # FP rate
    assert (nwords - len(nodups)) / nwords < 0.1
